@@ -47,7 +47,6 @@ from rasa.shared.nlu.constants import (
     ENTITIES,
 )
 
-
 if TYPE_CHECKING:
     from rasa.shared.core.trackers import DialogueStateTracker
 
@@ -229,15 +228,19 @@ class Domain:
     @classmethod
     def from_directory(cls, path: Text) -> "Domain":
         """Loads and merges multiple domain files recursively from a directory tree."""
-
-        domain = Domain.empty()
+        domain_dict = {}
         for root, _, files in os.walk(path, followlinks=True):
             for file in files:
                 full_path = os.path.join(root, file)
                 if Domain.is_domain_file(full_path):
-                    other = Domain.from_file(full_path)
-                    domain = other.merge(domain)
-
+                    _ = Domain.from_file(full_path)  # does the validation here only
+                    other_dict = rasa.shared.utils.io.read_yaml(
+                        rasa.shared.utils.io.read_file(full_path)
+                    )
+                    domain_dict = Domain.merge_domain_dicts(
+                        cls, domain_dict, other_dict
+                    )
+        domain = Domain.from_dict(domain_dict)
         return domain
 
     def merge(self, domain: Optional["Domain"], override: bool = False) -> "Domain":
@@ -245,8 +248,8 @@ class Domain:
 
         List attributes like ``intents`` and ``actions`` will be deduped
         and merged. Single attributes will be taken from `self` unless
-        override is `True`, in which case they are taken from `domain`."""
-
+        override is `True`, in which case they are taken from `domain`.
+        """
         if not domain or domain.is_empty():
             return self
 
@@ -256,34 +259,6 @@ class Domain:
         domain_dict = domain.as_dict()
         combined = self.as_dict()
 
-        def merge_dicts(
-            tempDict1: Dict[Text, Any],
-            tempDict2: Dict[Text, Any],
-            override_existing_values: bool = False,
-        ) -> Dict[Text, Any]:
-
-            if override_existing_values:
-                merge_dicts, b = tempDict1.copy(), tempDict2.copy()
-
-            else:
-                merge_dicts, b = tempDict2.copy(), tempDict1.copy()
-
-            merge_dicts.update(b)
-            return merge_dicts
-
-        def merge_lists(list1: List[Any], list2: List[Any]) -> List[Any]:
-            return sorted(list(set(list1 + list2)))
-
-        def merge_lists_of_dicts(
-            dict_list1: List[Dict],
-            dict_list2: List[Dict],
-            override_existing_values: bool = False,
-        ) -> List[Dict]:
-            dict1 = {list(i.keys())[0]: i for i in dict_list1}
-            dict2 = {list(i.keys())[0]: i for i in dict_list2}
-            merged_dicts = merge_dicts(dict1, dict2, override_existing_values)
-            return list(merged_dicts.values())
-
         if override:
             config = domain_dict["config"]
             for key, val in config.items():
@@ -292,9 +267,10 @@ class Domain:
         if override or self.session_config == SessionConfig.default():
             combined[SESSION_CONFIG_KEY] = domain_dict[SESSION_CONFIG_KEY]
 
-        combined[KEY_INTENTS] = merge_lists_of_dicts(
-            combined[KEY_INTENTS], domain_dict[KEY_INTENTS], override
-        )
+        if combined[KEY_INTENTS] or domain_dict[KEY_INTENTS]:
+            combined[KEY_INTENTS] = self.merge_lists_of_dicts(
+                combined[KEY_INTENTS], domain_dict[KEY_INTENTS], override
+            )
 
         # remove existing forms from new actions
         for form in combined[KEY_FORMS]:
@@ -302,15 +278,110 @@ class Domain:
                 domain_dict[KEY_ACTIONS].remove(form)
 
         for key in [KEY_ENTITIES, KEY_ACTIONS, KEY_E2E_ACTIONS]:
-            combined[key] = merge_lists(combined[key], domain_dict[key])
+            combined[key] = self.merge_lists(combined[key], domain_dict[key])
 
         for key in [KEY_FORMS, KEY_RESPONSES, KEY_SLOTS]:
-            combined[key] = merge_dicts(combined[key], domain_dict[key], override)
+            combined[key] = self.merge_dicts(combined[key], domain_dict[key], override)
 
         return self.__class__.from_dict(combined)
 
+    def merge_domain_dicts(
+        self, domain1: Dict, domain2: Dict, override: bool = False
+    ) -> Dict[Text, Any]:
+        """Merges this domain dict with another one, combining their attributes.
+
+        This is used when multiple domain yml files are configured in a single
+        directory. Unlike the merge method above, which merges Domain objects by
+        creating each object then merging it with the previous, this method merges
+        domain dicts, and ensures all attributes (like ``intents``, ``entities``, and
+        ``actions``) are known to the Domain when the object is created.
+
+        List attributes like ``intents`` and ``actions`` are deduped
+        and merged. Single attributes are taken from `domain1` unless
+        override is `True`, in which case they are taken from `domain2`.
+        """
+        if not domain2:
+            return domain1
+
+        if not domain1:
+            return domain2
+
+        domain_dict = domain2
+        combined = domain1
+
+        if override:
+            config = domain_dict["config"]
+            for key, val in config.items():
+                combined["config"][key] = val
+
+        if override or domain2.get("session_config"):
+            combined[SESSION_CONFIG_KEY] = domain_dict[SESSION_CONFIG_KEY]
+
+        if combined.get(KEY_INTENTS) or domain_dict.get(KEY_INTENTS):
+            combined[KEY_INTENTS] = combined.get(KEY_INTENTS, [])
+            domain_dict[KEY_INTENTS] = domain_dict.get(KEY_INTENTS, [])
+            combined[KEY_INTENTS] = self.merge_lists_of_dicts(
+                combined[KEY_INTENTS], domain_dict[KEY_INTENTS], override
+            )
+
+        # remove existing forms from new actions
+        for form in combined.get(KEY_FORMS, []):
+            if form in domain_dict.get(KEY_ACTIONS, []):
+                domain_dict[KEY_ACTIONS].remove(form)
+
+        for key in [KEY_ENTITIES, KEY_ACTIONS, KEY_E2E_ACTIONS]:
+            combined[key] = self.merge_lists(
+                combined.get(key, []), domain_dict.get(key, [])
+            )
+
+        for key in [KEY_FORMS, KEY_RESPONSES, KEY_SLOTS]:
+            combined[key] = self.merge_dicts(
+                combined.get(key, {}), domain_dict.get(key, {}), override
+            )
+
+        return combined
+
+    @staticmethod
+    def merge_dicts(
+        tempDict1: Dict[Text, Any],
+        tempDict2: Dict[Text, Any],
+        override_existing_values: bool = False,
+    ) -> Dict[Text, Any]:
+        """Merges 2 dicts."""
+        if override_existing_values:
+            merged_dicts, b = tempDict1.copy(), tempDict2.copy()
+
+        else:
+            merged_dicts, b = tempDict2.copy(), tempDict1.copy()
+        merged_dicts.update(b)
+        return merged_dicts
+
+    @staticmethod
+    def merge_lists(list1: List[Any], list2: List[Any]) -> List[Any]:
+        """Merges 2 lists."""
+        return sorted(list(set(list1 + list2)))
+
+    @staticmethod
+    def merge_lists_of_dicts(
+        dict_list1: List[Dict],
+        dict_list2: List[Dict],
+        override_existing_values: bool = False,
+    ) -> List[Dict]:
+        """Merges 2 dict lists."""
+        dict1 = {
+            (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i
+            for i in dict_list1
+        }
+        dict2 = {
+            (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i
+            for i in dict_list2
+        }
+        merged_dicts = Domain.merge_dicts(dict1, dict2, override_existing_values)
+        return list(merged_dicts.values())
+
     @staticmethod
     def collect_slots(slot_dict: Dict[Text, Any]) -> List[Slot]:
+        """Collects the slots."""
         slots = []
         # make a copy to not alter the input dictionary
         slot_dict = copy.deepcopy(slot_dict)
@@ -1460,9 +1531,10 @@ class Domain:
             The intent properties as they are used in domain files.
         """
         intent_properties = copy.deepcopy(self.intent_properties)
+        sorted_intent_properties = sorted(intent_properties.items())
         intents_for_file = []
 
-        for intent_name, intent_props in intent_properties.items():
+        for intent_name, intent_props in sorted_intent_properties:
             if (
                 intent_name in rasa.shared.core.constants.DEFAULT_INTENTS
                 and intent_name not in self.overridden_default_intents
@@ -1481,12 +1553,15 @@ class Domain:
             if len(use_entities) == len(self.entities):
                 intent_props[USE_ENTITIES_KEY] = True
             elif len(use_entities) <= len(self.entities) / 2:
-                intent_props[USE_ENTITIES_KEY] = list(use_entities)
+                entities = list(use_entities)
+                entities.sort()
+                intent_props[USE_ENTITIES_KEY] = entities
             else:
-                intent_props[IGNORE_ENTITIES_KEY] = list(ignore_entities)
+                entities = list(ignore_entities)
+                entities.sort()
+                intent_props[IGNORE_ENTITIES_KEY] = entities
             intent_props.pop(USED_ENTITIES_KEY)
             intents_for_file.append({intent_name: intent_props})
-
         return intents_for_file
 
     def _transform_entities_for_file(self) -> List[Union[Text, Dict[Text, Any]]]:
